import abc
from typing import (
    List,
    TypeVar,
    Type,
    TYPE_CHECKING,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
    from xoa_driver.internals.utils.indices import observer


CT = TypeVar("CT")


class BaseIndex(abc.ABC):
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "observer.IndicesObserver") -> None:
        self._observer = observer
        self._conn = conn
        self.kind = kind

    @property
    def idx(self) -> int:
        return self.kind.index_id

    @abc.abstractmethod
    async def delete(self) -> None:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    async def _fetch(cls, conn: "itf.IConnection", module_id: int, port_id: int) -> List[int]:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    async def _new(cls: Type[CT], conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "observer.IndicesObserver") -> CT:
        raise NotImplementedError()



from typing import (
    TYPE_CHECKING,
    List,
    TypeVar,
    Type,
)
from xoa_driver.internals.core.commands import (
    PL_INDICES,
    PL_CREATE,
    PL_DELETE,
    PL_LENGTH,
)

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
from xoa_driver.internals.utils.indices import observer as idx_obs
from .base_index import BaseIndex


LT = TypeVar("LT")


class LengthTermIdx(BaseIndex):
    """L23 Length Term Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)
        self.length = PL_LENGTH(conn, *kind)
        """Configure length term
        :type: PL_LENGTH
        """

    async def delete(self):
        await PL_DELETE(self._conn, *self.kind).set()
        """Delete length term.
        :type: PL_DELETE
        """

        self._observer.notify(idx_obs.IndexEvents.DEL, self)

    @classmethod
    async def _fetch(cls, conn: "itf.IConnection", module_id: int, port_id: int) -> List[int]:
        resp = await PL_INDICES(conn, module_id, port_id).get()
        return list(resp.length_term_xindices)

    @classmethod
    async def _new(cls: Type[LT], conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> LT:
        await PL_CREATE(conn, *kind).set()
        return cls(conn, kind, observer)



import typing
from xoa_driver import ports
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.utils import ports_manager as pm

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import __interfaces as m_itf

from .module_l23_base import ModuleL23


@typing.final
@revisions.register_valkyrie_module(rev="Thor-400G-7S-1P")
class MThor400G7S1P(ModuleL23):
    """Test module Thor-400G-7S-1P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PThor400G7S1P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PThor400G7S1P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index of Thor-400G-7S-1P"""



from typing import TYPE_CHECKING
from xoa_driver.internals.core.commands import (
    P_FAULTSIGNALING,
    P_FAULTSTATUS,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf


class Fault:
    """L23 port fault settings."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.signaling = P_FAULTSIGNALING(conn, module_id, port_id)
        """L23 port fault signaling.
        Representation ofP_FAULTSIGNALING
        """

        self.status = P_FAULTSTATUS(conn, module_id, port_id)
        """L23 port fault status.
        Representation of P_FAULTSTATUS
        """



from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    C_IPADDRESS,
    C_DHCP,
    C_MACADDRESS,
    C_HOSTNAME,
)


class ManagementInterface:
    """
    Tester management interface address configuration.
    """

    def __init__(self, conn: "itf.IConnection") -> None:
        self.ip_address = C_IPADDRESS(conn)
        """
        Chassis management port IP address configuration.
        
        :type: C_IPADDRESS
        """

        self.dhcp = C_DHCP(conn)
        """
        Chassis management port IP address configuration with DHCP.
                
        :type: C_DHCP
        """

        self.macaddress = C_MACADDRESS(conn)
        """
        Chassis management port MAC address.
                
        :type: C_MACADDRESS
        """

        self.hostname = C_HOSTNAME(conn)
        """
        Chassis hostname.
                
        :type: C_HOSTNAME
        """



from typing import (
    Dict,
    Type,
)
from .interfaces import CMD_TYPE

COMMANDS_REGISTRY: Dict[int, Type[CMD_TYPE]] = {}
"""A global registry of which map command id to its class implementation."""


def register_command(xmc_cls: Type[CMD_TYPE]) -> Type:
    """A decorator used for register commands descriptors in to global registry"""
    global COMMANDS_REGISTRY
    if xmc_cls.code in COMMANDS_REGISTRY:
        raise RuntimeError(
            f"Command with the same code {xmc_cls.code} for the command {COMMANDS_REGISTRY[xmc_cls.code]}!"
        )
    COMMANDS_REGISTRY[xmc_cls.code] = xmc_cls
    return xmc_cls



"""
L47 Module Commands
"""
from dataclasses import dataclass
import typing
import functools

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403


@register_command
@dataclass
class M4_SYSTEMID:
    """
    Return the system identifier of a L47 module.
    """

    code: typing.ClassVar[int] = 803
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        system_id: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, module system identifier.

    def get(self) -> "Token[GetDataAttr]":
        """Get the system identifier of a L47 module.

        :return: the system identifier of a L47 module.
        :rtype: M4_SYSTEMID.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_VERSIONNO:
    """
    Returns a version string containing a combination of information regarding the
    software version and the build environment. The first part of the string is the
    software build version.
    """

    code: typing.ClassVar[int] = 804
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        version_string: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, module version string.

    def get(self) -> "Token[GetDataAttr]":
        """Get a version string containing a combination of information regarding the software version and the build environment.

        :return: a version string containing a combination of information regarding the software version and the build environment
        :rtype: M4_VERSIONNO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_SYSTEM_STATUS:
    """
    Returns the L47 module system status in a text string.
    """

    code: typing.ClassVar[int] = 805
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        status_string: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, module status string.

    def get(self) -> "Token[GetDataAttr]":
        """Get the L47 module system status in a text string

        :return: the L47 module system status in a text string
        :rtype: M4_SYSTEM_STATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_COMPATIBLE_CLIENT_VERSION:
    """
    Returns the recommended and required VulcanMananger client version.
    """

    code: typing.ClassVar[int] = 806
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        recommended_major: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, recommended major version.
        recommended_minor: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, recommended minor version.
        recommended_minor_2: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, recommended minor 2 version.
        recommended_minor_3: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, recommended minor 3 version.
        required_major: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, required major version.
        required_minor: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, required minor version.
        required_minor_2: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, required minor 2 version.
        required_minor_3: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # string, required minor 3 version.

    def get(self) -> "Token[GetDataAttr]":
        """Get the recommended and required VulcanMananger client version.

        :return: the recommended and required VulcanMananger client version.
        :rtype: M4_COMPATIBLE_CLIENT_VERSION.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_TIME:
    """
    Returns the module time in millisecond.
    """

    code: typing.ClassVar[int] = 807
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        time_now: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, the current time (mSec since module restart)

    def get(self) -> "Token[GetDataAttr]":
        """Get the module time in millisecond.

        :return: the module time in millisecond.
        :rtype: M4_TIME.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_SYSTEM_TIME:
    """
    Sets or returns the modules system time in UTC.
    """

    code: typing.ClassVar[int] = 808
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        year: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the year.
        month: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the month.
        day: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the day of the mont.
        hour: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the hour.
        minute: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the minute.
        second: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the second.

    @dataclass(frozen=True)
    class GetDataAttr:
        year: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the year.
        month: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the month.
        day: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the day of the mont.
        hour: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the hour.
        minute: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the minute.
        second: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the second.

    def get(self) -> "Token[GetDataAttr]":
        """Get the modules system time in UTC.

        :return: the modules system time in UTC.
        :rtype: M4_SYSTEM_TIME.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, year: int, month: int, day: int, hour: int, minute: int, second: int) -> "Token":
        """Set the modules system time in UTC.

        :param year: the year
        :type year: int
        :param month: the month
        :type month: int
        :param day: the day of the month
        :type day: int
        :param hour: the hour
        :type hour: int
        :param minute: the minute
        :type minute: int
        :param second: the second
        :type second: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, year=year, month=month, day=day, hour=hour, minute=minute, second=second))


@register_command
@dataclass
class M4_MEM_INFO:
    """
    Return the system memory information.
    """

    code: typing.ClassVar[int] = 809
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        year: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, total memory.
        month: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, free memory.

    def get(self) -> "Token[GetDataAttr]":
        """Get the system memory information.

        :return: the system memory information.
        :rtype: M4_MEM_INFO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_CAPTURE_SIZE:
    """
    Specify whether to capture whole packets(large) or truncated packets. When
    truncated (small) is selected only the first 128 bytes of the packet are saved.
    """

    code: typing.ClassVar[int] = 810
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        size: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=CaptureSize)  # coded byte, specifying whether to capture whole packets or truncated packets.

    @dataclass(frozen=True)
    class GetDataAttr:
        size: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=CaptureSize)  # coded byte, specifying whether to capture whole packets or truncated packets.

    def get(self) -> "Token[GetDataAttr]":
        """Get whether to capture whole packets(large) or truncated packets.

        :return: whether to capture whole packets(large) or truncated packets.
        :rtype: M4_CAPTURE_SIZE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, size: CaptureSize) -> "Token":
        """Set whether to capture whole packets(large) or truncated packets.

        :param size: specifying whether to capture whole packets or truncated packets.
        :type size: CaptureSize
        """
        return Token(self._connection, build_set_request(self, module=self._module, size=size))

    set_full = functools.partialmethod(set, CaptureSize.FULL)
    """Capture whole packets"""
    set_small = functools.partialmethod(set, CaptureSize.SMALL)
    """Capture truncated packets"""


@register_command
@dataclass
class M4_LICENSE_INFO:
    """
    Returns the number of available and free PE licenses. Only 'available' number of PEs
    can simultaneously be assigned to reserved ports.
    """

    code: typing.ClassVar[int] = 820
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        pes_available: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, number of PEs that are licensed on the module, and can be used simultaneously.
        pes_free: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, number of free PE licenses on the module 1G available: integer, number of 1G licenses on the module, that can be used simultaneously.
        N1g_available: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, number of 1G speed licenses that are licensed on the module, and can be used simultaneously.
        N1g_free: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, number of 1G speed licenses on the module 1G available: integer, number of 1G licenses on the module, that can be used simultaneously.
        N10g_available: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, number of 10G speed licenses that are licensed on the module, and can be used simultaneously.
        N10g_free: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, number of free 10G speed licenses on the module 1G available: integer, number of 1G licenses on the module, that can be used simultaneously.
        N25g_available: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, number of 25G speed licenses that are licensed on the module, and can be used simultaneously.
        N25g_free: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, number of free 25G speed licenses on the module 1G available: integer, number of 1G licenses on the module, that can be used simultaneously.
        N40g_available: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, number of 40G speed licenses that are licensed on the module, and can be used simultaneously.
        N40g_free: XmpField[xt.XmpInt] = XmpField(
            xt.XmpInt
        )  # integer, number of free 40G speed licenses on the module 1G available: integer, number of 1G licenses on the module, that can be used simultaneously.

    def get(self) -> "Token[GetDataAttr]":
        """Get the number of available and free PE licenses.

        :return: the number of available and free PE licenses
        :rtype: M4_LICENSE_INFO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_REPLAY_PARSE_START:
    """
    Command to start parsing an uploaded Capture File (in PCAP format) intended for
    use in a replay test scenario. The result of the parsing - if successful - is a
    Replay File (in BSON format) with the same name as the Capture File, which can
    be used as parameter to P4G_REPLAY_filename command. If parsing is unsuccessful,
    a Replay File is created containing the parse result. The
    M4_REPLAY_FILE_INFO_BSON command can be used to get information about a Replay
    File - including the parse result. PCAP Capture Files can be uploaded to the L47
    chassis using FTP. The 'root' location of Capture Files uploaded manually by the
    user is /var/ftp/pub/replay/pcap/. Three subdirectories exist: cache/, user/ and
    xena/. cache / and xena/ is used by Vulcan Manager, and user/ is intended for
    manual upload and parsing of Capture Files. A similar directory structure is
    present for Replay Files generated by the parsing, and the 'root' location is
    /var/ftp/pub/replay/bson/.
    """

    code: typing.ClassVar[int] = 830
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        filename: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, filename (including relative path and excluding the '.pcap' extension).

    def set(self, filename: str) -> "Token":
        """Start parsing an uploaded Capture File

        :param filename: filename (including relative path and excluding the '.pcap' extension).
        :type filename: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, filename=filename))


@register_command
@dataclass
class M4_REPLAY_PARSE_STOP:
    """
    Command to stop parsing a Capture File. Parsing of very large Capture Files may
    take several seconds, and may be aborted using this command. No parameters
    """

    code: typing.ClassVar[int] = 831
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Stop parsing a Capture File.
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
            ),
        )


@register_command
@dataclass
class M4_REPLAY_PARSE_STATE:
    """
    Only one Capture File can be parsed at a time. This command returns the state of
    the parser, which can be PARSING or OFF. M4_REPLAY_PARSE_START command is only
    accepted when the parser state is OFF.
    """

    code: typing.ClassVar[int] = 832
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        state: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=ReplayParserState)  # coded byte, state of the replay parser

    def get(self) -> "Token[GetDataAttr]":
        """Get capture file parsing state.

        :return: capture file parsing state
        :rtype: M4_REPLAY_PARSE_STATE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_REPLAY_PARSER_PARAMS:
    """
    Configuration of parameters for the parsing of pcap files.
    """

    code: typing.ClassVar[int] = 833
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        tcp_port: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, server TCP Port of dummy TCP connection inserted in UDP only replay files

    @dataclass(frozen=True)
    class GetDataAttr:
        tcp_port: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, server TCP Port of dummy TCP connection inserted in UDP only replay files

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of parameters for the parsing of pcap files.

        :return: the configuration of parameters for the parsing of pcap files
        :rtype: M4_REPLAY_PARSER_PARAMS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, tcp_port: int) -> "Token":
        """Set the configuration of parameters for the parsing of pcap files.

        :param tcp_port: server-side TCP port of the dummy TCP connection inserted in UDP.
        :type tcp_port: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, tcp_port=tcp_port))


@register_command
@dataclass
class M4_REPLAY_FILE_LIST_BSON:
    """
    Works as ``M4_REPLAY_FILE_LIST``, but returns the file list formatted as a BSON
    document.
    """

    code: typing.ClassVar[int] = 840
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        bson: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # list of hex bytes, bson document containing the file list

    def get(self) -> "Token[GetDataAttr]":
        """Get the replay file list in BSON document format.

        :return: the replay file list in BSON format
        :rtype: M4_REPLAY_FILE_LIST_BSON.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_REPLAY_FILE_LIST:
    """
    Returns a list of Replay Files (``.bson`` files) in the 'user' Replay File
    directory (``/var/ftp/pub/replay/bson/user/``).
    """

    code: typing.ClassVar[int] = 841
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        file_list: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, comma separated list of filenames excluding the '.bson' extension.

    def get(self) -> "Token[GetDataAttr]":
        """Generate a list of Replay Files in BSON document on the tester.

        :return: a list of Replay Files in BSON document on the tester
        :rtype: M4_REPLAY_FILE_LIST.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_CAPTURE_FILE_LIST_BSON:
    """
    Works as ``M4_CAPTURE_FILE_LIST``, but returns the file list formatted as a BSON
    document.
    """

    code: typing.ClassVar[int] = 842
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        bson: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # list of hex bytes, bson document containing the file list

    def get(self) -> "Token[GetDataAttr]":
        """Get the capture file list in BSON document.

        :return: the capture file list in BSON document
        :rtype: M4_CAPTURE_FILE_LIST_BSON.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_CAPTURE_FILE_LIST:
    """
    Returns a list of Capture Files (``.pcap`` files) in the 'user' Capture File
    directory (``/var/ftp/pub/replay/pcap/user/``).
    """

    code: typing.ClassVar[int] = 843
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        file_list: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, comma separated list of filenames excluding the '.pcap' extension.

    def get(self) -> "Token[GetDataAttr]":
        """Generate a list of Capture Files in BSON document on the tester.

        :return: list of Capture Files in BSON document on the tester
        :rtype: M4_CAPTURE_FILE_LIST.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))


@register_command
@dataclass
class M4_REPLAY_FILE_DELETE:
    """
    Command to delete a Replay File (``.bson`` file) in the Replay File directory
    (``/var/ftp/pub/replay/bson/``). For information about the location and directory
    structure for the Replay Files, see: M4_REPLAY_PARSE_START
    """

    code: typing.ClassVar[int] = 845
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        filename: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, file name (including relative path and excluding the '.bson' extension).

    def set(self, filename: str) -> "Token":
        """Delete a Replay File in the Replay File directory.

        :param filename: file name (including relative path and excluding the ``.bson`` extension).
        :type filename: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, filename=filename))


@register_command
@dataclass
class M4_CAPTURE_FILE_DELETE:
    """
    Command to delete a Capture File (``.pcap`` file) in the Capture File directory
    (``/var/ftp/pub/replay/pcap/``). For information about the location and directory
    structure for the Capture Files, see: M4_REPLAY_PARSE_START
    """

    code: typing.ClassVar[int] = 846
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        filename: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, file name (including relative path and excluding the '.pcap' extension).

    def set(self, filename: str) -> "Token":
        """Delete a Capture File in the Capture File directory.

        :param filename: file name (including relative path and excluding the ``.pcap`` extension).
        :type filename: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, filename=filename))


@register_command
@dataclass
class M4_TLS_CIPHER_SUITES:
    """
    Returns a list of supported TLS Cipher Suites.
    """

    code: typing.ClassVar[int] = 852
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class GetDataAttr:
        cipher_suites: XmpField[xt.XmpHexList] = XmpField(xt.XmpHexList)  # list of hex bytes, list of IANA values of supported cipher suites

    def get(self) -> "Token[GetDataAttr]":
        """Get a list of supported TLS Cipher Suites.

        :return: list of IANA values of supported cipher suites
        :rtype: M4_TLS_CIPHER_SUITES.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))



from typing import Type
from xoa_driver.internals.core.commands import (
    C_PORTCOUNTS,
    C_MULTIUSER,
    C_TRAFFIC,
    C_TRAFFICSYNC,
    C_VERSIONNO_MINOR,
)
from xoa_driver.internals.utils.modules_manager import ModulesManager
from ._base_tester import BaseTester


from xoa_driver.v2 import modules
from xoa_driver.internals.state_storage import testers_state


TypeL23Manager = ModulesManager["modules.ModuleL23VE"]


def get_module_type(revision: str) -> Type:
    return modules.ModuleL23VE


class L23VeTester(BaseTester["testers_state.TesterLocalState"]):
    """
    Representation of a virtual Xena Valkyrie Tester.
    """

    def __init__(self, host: str, username: str, password: str = "xena", port: int = 22606, *, debug: bool = False) -> None:
        super().__init__(host=host, username=username, password=password, port=port, debug=debug)

        self._local_states = testers_state.TesterLocalState(host, port)

        self.multiuser = C_MULTIUSER(self._conn)
        """
        Representation of C_MULTIUSER
        """

        self.traffic = C_TRAFFIC(self._conn)
        """
        Representation of C_TRAFFIC
        """

        self.traffic_sync = C_TRAFFICSYNC(self._conn)
        """
        Representation of C_TRAFFICSYNC
        """

        self.version_no_minor = C_VERSIONNO_MINOR(self._conn)
        """
        Representation of C_VERSIONNO_MINOR
        """

        self.modules: TypeL23Manager = ModulesManager(self._conn, get_module_type)
        """
        Module index manager of the L23 tester.
        """

    @property
    def info(self) -> testers_state.TesterLocalState:
        return self._local_states

    async def _setup(self):
        await super()._setup()
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)

        ft_pc = await C_PORTCOUNTS(self._conn).get()
        port_counts = ft_pc.port_counts
        await self.modules.fill_l23(port_counts)
        return self



import typing
if typing.TYPE_CHECKING:
    from .protocol.struct_request import Request
    from asyncio import Future


class IsDataclass(typing.Protocol):
    __dataclass_fields__: typing.Dict[str, typing.Any]
    def __init__(self, *args, **kwargs) -> None: ...  # noqa: E704


DC = typing.TypeVar("DC", bound=IsDataclass)


class ICommand(typing.Protocol):
    code: typing.ClassVar[int]
    pushed: typing.ClassVar[bool]


class ICmdOnlySet(ICommand, typing.Protocol):
    """A template class which provide only <cmd_set> method."""
    SetDataAttr: typing.Type[DC]  # type: ignore
    set: typing.Callable


class ICmdOnlyGet(ICommand, typing.Protocol):
    """A template class which provide only <cmd_get> method."""
    GetDataAttr: typing.Type[DC]  # type: ignore
    get: typing.Callable


CMD_TYPE = typing.Union[ICmdOnlySet, ICmdOnlyGet]


Inst = typing.TypeVar('Inst')
CallbackType = typing.Callable[[Inst, typing.Optional[IsDataclass]], typing.Awaitable[None]]


class IConnection(typing.Protocol):
    """Representation of TransportationHandler"""
    is_connected: bool

    def send(self, requests: bytes) -> None:
        ...

    def close(self) -> None:
        ...

    async def prepare_data(self, request: "Request") -> typing.Tuple[bytes, "Future"]:
        ...

    def subscribe(self, xmc_cls: "CMD_TYPE", callback: "typing.Callable") -> None:
        ...



from typing import (
    Generic,
    TypeVar,
    Tuple,
)
from collections import OrderedDict


T = TypeVar("T")


class ResourcesBaseManager(Generic[T]):
    """Resource Base Manager
    """
    __slots__ = ("_items", "_lock", )

    def __init__(self) -> None:
        self._items: OrderedDict[int, T] = OrderedDict()
        self._lock: bool = False

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self):
        return iter(self._items.values())

    def obtain(self, key: int) -> T:
        """Obtain a single resource.
        """
        return self._items[key]

    def obtain_multiple(self, *keys: int) -> Tuple[T, ...]:
        """Obtain multiple resources as a tuple of resources.
        """
        return tuple(self._items[k] for k in keys)




from xoa_driver.internals.core.commands import (
    C_BUILDSTRING,
    C_VERSIONNO_MINOR
)
from .l47_tester import L47Tester


class L47VeTester(L47Tester):
    """
    Representation of a virtual Xena Vulcan Tester.
    """

    def __init__(self, host: str, username: str, password: str = "xena", port: int = 22606, *, debug: bool = False) -> None:
        super().__init__(host=host, username=username, password=password, port=port, debug=debug)
        self.version_no_minor = C_VERSIONNO_MINOR(self._conn)
        """
        Representation of C_VERSIONNO_MINOR
        """

        self.build_string = C_BUILDSTRING(self._conn)
        """
        Representation of C_BUILDSTRING
        """



from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from . import struct_header
from . import struct_response
from . import struct_request
from .constants import (
    CommandType,
    NOTHING,
)
from xoa_driver.internals.core.interfaces import (
    ICmdOnlySet,
    ICmdOnlyGet,
    CMD_TYPE,
)


def build_set_request(cls: ICmdOnlySet, **kwargs) -> "struct_request.Request":
    indices = kwargs.pop("indices", [])
    module = kwargs.pop("module", NOTHING)
    port = kwargs.pop("port", NOTHING)
    req_values = cls.SetDataAttr(**kwargs)
    set_get = CommandType.COMMAND_VALUE.value
    return struct_request.Request(
        type(cls).__name__,
        indices,
        set_get,
        cls.code,
        module,
        port,
        req_values
    )


def build_get_request(cls: ICmdOnlyGet, **kwargs) -> "struct_request.Request":
    indices = kwargs.pop("indices", [])
    module = kwargs.pop("module", NOTHING)
    port = kwargs.pop("port", NOTHING)
    req_values = None
    set_get = CommandType.COMMAND_QUERY.value
    return struct_request.Request(
        type(cls).__name__,
        indices,
        set_get,
        cls.code,
        module,
        port,
        req_values
    )


def build_from_bytes(cls: Type[CMD_TYPE], header: "struct_header.ResponseHeader", data: bytes) -> "struct_response.Response":
    """Parse bytes retrieved from server to Response structure."""
    properties_structure = getattr(cls, "GetDataAttr", None)
    return struct_response.Response(header, cls.__name__, data, properties_structure)



from typing import (
    Dict,
    Tuple,
    Type,
    TypeVar,
    TYPE_CHECKING
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands.enums import (
    FilterType,
    FilterVlanType,
    ProtocolOption,
    FilterMode,
)
from xoa_driver.internals.core.commands import (
    PEF_ETHSETTINGS,
    PEF_ETHSRCADDR,
    PEF_ETHDESTADDR,
    PEF_L2PUSE,
    PEF_VLANSETTINGS,
    PEF_VLANTAG,
    PEF_VLANPCP,
    PEF_MPLSSETTINGS,
    PEF_MPLSLABEL,
    PEF_MPLSTOC,
    PEF_L3USE,
    PEF_IPV4SETTINGS,
    PEF_IPV4SRCADDR,
    PEF_IPV4DESTADDR,
    PEF_IPV4DSCP,
    PEF_IPV6SETTINGS,
    PEF_IPV6SRCADDR,
    PEF_IPV6DESTADDR,
    PEF_IPV6TC,
    PEF_UDPSETTINGS,
    PEF_UDPSRCPORT,
    PEF_UDPDESTPORT,
    PEF_TCPSETTINGS,
    PEF_TCPSRCPORT,
    PEF_TCPDESTPORT,
    PEF_ANYSETTINGS,
    PEF_ANYCONFIG,
    PEF_TPLDSETTINGS,
    PEF_TPLDCONFIG,
    # PEF_ISSHADOWDIRTY, #TODO: need to implement
    # PEF_CANCEL, #TODO: need to implement
    PEF_VALUE,
    PEF_MASK,
    PEF_PROTOCOL,
    PEF_MODE,
)
from ._utils import prevent_set


class FEthernet:
    """Filter for Ethernet field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_ETHSETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on Ethernet field.
        Representation of PEF_ETHSETTINGS
        """
        self.src_address = prevent_set(PEF_ETHSRCADDR(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Ethernet source address settings for the filter.
        Representation of PEF_ETHSRCADDR
        """
        self.dest_address = prevent_set(PEF_ETHDESTADDR(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Ethernet destination address settings for the filter.
        Representation of PEF_ETHDESTADDR
        """


class FPerVlanType:
    """Filter for VLAN tag."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType, vlan_type: FilterVlanType) -> None:
        self.tag = prevent_set(PEF_VLANTAG(conn, module_id, port_id, flow_index, filter_type, vlan_type), filter_type)
        """VLAN tag setting for the filter.
        Representation of PEF_VLANTAG
        """
        self.pcp = prevent_set(PEF_VLANPCP(conn, module_id, port_id, flow_index, filter_type, vlan_type), filter_type)
        """VLAN PCP setting for the filter.
        Representation of PEF_VLANPCP
        """


class FVlan:
    """Filter for VLAN field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_VLANSETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on VLAN field.
        Representation of PEF_VLANSETTINGS
        """

        self.inner = FPerVlanType(conn, module_id, port_id, flow_index, filter_type, FilterVlanType.INNER)
        """Filter for inner VLAN tag."""

        self.outer = FPerVlanType(conn, module_id, port_id, flow_index, filter_type, FilterVlanType.OUTER)
        """Filter for outer VLAN tag."""


class FUdp:
    """Filer for UDP field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_UDPSETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on UDP field.
        Representation of PEF_UDPSETTINGS
        """
        self.src_port = prevent_set(PEF_UDPSRCPORT(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """UDP source port settings for the filter.
        Representation of PEF_UDPSRCPORT
        """
        self.dest_port = prevent_set(PEF_UDPDESTPORT(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """UDP destination port settings for the filter.
        Representation of PEF_UDPDESTPORT
        """


class FTcp:
    """Filter for TCP."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_TCPSETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on TCP field.
        Representation of PEF_TCPSETTINGS
        """
        self.src_port = prevent_set(PEF_TCPSRCPORT(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """TCP source port settings for the filter.
        Representation of PEF_TCPSRCPORT
        """
        self.dest_port = prevent_set(PEF_TCPDESTPORT(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """TCP destination port settings for the filter.
        Representation of PEF_TCPDESTPORT
        """


class FIPv4:
    """Filter for IPv4 field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_IPV4SETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on IPv4 field.
        Representation of PEF_IPV4SETTINGS
        """
        self.src_address = prevent_set(PEF_IPV4SRCADDR(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """IPv4 source address settings for the filter.
        Representation of PEF_IPV4SRCADDR
        """
        self.dest_address = prevent_set(PEF_IPV4DESTADDR(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """IPv4 destination address settings for the filter.
        Representation of PEF_IPV4DESTADDR
        """
        self.dscp = prevent_set(PEF_IPV4DSCP(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """IPv4 DSCP/TOS settings for the filter.
        Representation of PEF_IPV4DSCP
        """


class FIPv6:
    """Filter for IPv6 field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_IPV6SETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on IPv6 field.
        Representation of PEF_IPV6SETTINGS
        """
        self.src_address = prevent_set(PEF_IPV6SRCADDR(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """IPv6 source address settings for the filter.
        Representation of PEF_IPV6SRCADDR
        """
        self.dest_address = prevent_set(PEF_IPV6DESTADDR(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """IPv6 destination address settings for the filter.
        Representation of PEF_IPV6DESTADDR
        """
        self.traffic_class = prevent_set(PEF_IPV6TC(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """IPv6 traffic class settings for the filter.
        Representation of PEF_IPV6TC
        """


class FIp:
    """Filter for IP field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.v4 = FIPv4(conn, module_id, port_id, flow_index, filter_type)
        """Filter for IPv4 field."""

        self.v6 = FIPv6(conn, module_id, port_id, flow_index, filter_type)
        """Filter for IPv6 field."""


class FMpls:
    """Filter for MPLS field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_MPLSSETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on MPLS field.
        Representation of PEF_MPLSSETTINGS
        """
        self.label = prevent_set(PEF_MPLSLABEL(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """MPLS label settings for the filter.
        Representation of PEF_MPLSLABEL
        """
        self.toc = prevent_set(PEF_MPLSTOC(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """"MPLS TOC settings for the filter.
        Representation of PEF_MPLSTOC
        """


class FTpld:
    """Filter for Xena TPLD field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_TPLDSETTINGS(conn, module_id, port_id, flow_index, filter_type))
        """Filter action on Xena TPLD field.
        Representation of PEF_TPLDSETTINGS
        """

        self.test_payload_filters_config = tuple(
            prevent_set(PEF_TPLDCONFIG(conn, module_id, port_id, flow_index, filter_type, test_payload_filter_index), filter_type)
            for test_payload_filter_index in range(16)  # range need to put to the constants place
        )
        """TPLD filter configurations.
        Representation of PEF_TPLDCONFIG
        """


class FAny:
    """Filter for any field."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.settings = prevent_set(PEF_ANYSETTINGS(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Filter action on any field.
        Representation of PEF_ANYSETTINGS
        """
        self.config = prevent_set(PEF_ANYCONFIG(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """Any field's filter configurations.
        Representation of PEF_ANYCONFIG
        """


class ModeBasic:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self.l2plus_use = prevent_set(PEF_L2PUSE(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """L2 protocol to use.
        Representation of PEF_L2PUSE
        """
        self.l3_use = prevent_set(PEF_L3USE(conn, module_id, port_id, flow_index, filter_type), filter_type)
        """L3 protocol to use.
        Representation of PEF_L3USE
        """
        self.any = FAny(conn, module_id, port_id, flow_index, filter_type)
        """Filter for any field."""

        self.tpld = FTpld(conn, module_id, port_id, flow_index, filter_type)
        """Filter for Xena TPLD field."""

        self.mpls = FMpls(conn, module_id, port_id, flow_index, filter_type)
        """Filter for MPLS field."""

        self.ip = FIp(conn, module_id, port_id, flow_index, filter_type)
        """Filter for IP field."""

        self.tcp = FTcp(conn, module_id, port_id, flow_index, filter_type)
        """Filter for TCP field."""

        self.udp = FUdp(conn, module_id, port_id, flow_index, filter_type)
        """Filter for UDP field."""

        self.vlan = FVlan(conn, module_id, port_id, flow_index, filter_type)
        """Filter for VLAN field."""

        self.ethernet = FEthernet(conn, module_id, port_id, flow_index, filter_type)
        """Filter for Ethernet field."""


class ProtocolSegment:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType, protocol_segment_index: int, segment_type: ProtocolOption) -> None:  # noqa: E501
        self.segment_type = segment_type

        self.value = prevent_set(PEF_VALUE(conn, module_id, port_id, flow_index, filter_type, protocol_segment_index), filter_type)
        """Value bytes match for the filter.
        Representation of PEF_VALUE
        """
        self.mask = prevent_set(PEF_MASK(conn, module_id, port_id, flow_index, filter_type, protocol_segment_index), filter_type)
        """Mask byte value.
        Representation of PEF_MASK
        """


class ModeExtended:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType) -> None:
        self._conn = conn
        self._module_id = module_id
        self._port_id = port_id
        self._flow_index = flow_index
        self._filter_type = filter_type

        # self.protocol = prevent_set(PEF_PROTOCOL(conn, module_id, port_id, flow_index, filter_type))
        # """Protocol segments match for the filter.
        # Representation of PEF_PROTOCOL
        # """

    async def get_protocol_segments(self) -> Tuple[ProtocolSegment, ...]:
        segments_raw = (
            await PEF_PROTOCOL(
                self._conn,
                self._module_id,
                self._port_id,
                self._flow_index,
                self._filter_type
            ).get()
        ).segment_list
        return tuple(
            ProtocolSegment(
                self._conn,
                self._module_id,
                self._port_id,
                self._flow_index,
                self._filter_type,
                idx + 1,  # TODO: need to discuss with Leo
                ProtocolOption(segment_type)
            )
            for idx, segment_type in enumerate(segments_raw)
        )


T = TypeVar("T")


async def get_mode(conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int, filter_type: FilterType, mode_types: Dict[FilterMode, Type[T]]) -> T:
    mode_raw = (
        await PEF_MODE(
            conn,
            module_id,
            port_id,
            flow_index,
            filter_type,
        ).get()
    ).mode
    mode_type = mode_types[FilterMode(mode_raw)]
    return mode_type(
        conn,
        module_id,
        port_id,
        flow_index,
        filter_type,
    )



from __future__ import annotations

from xoa_driver.ports import GenericAnyPort
from xoa_driver.modules import GenericAnyModule

class ConfigError(Exception):
    msg: str


class NotConnectedError(ConfigError):
    def __init__(self) -> None:
        self.msg = "No tester is connected!"
        super().__init__(self.msg)


class NoSuchModuleError(ConfigError):
    def __init__(self, module_id: int) -> None:
        self.msg = f"No such module {module_id}!"
        super().__init__(self.msg)


class NoSuchPortError(ConfigError):
    def __init__(self, port_id: int) -> None:
        self.msg = f"No such port {port_id}!"
        super().__init__(self.msg)


class NotSupportPcsPmaError(ConfigError):
    def __init__(self, port: GenericAnyPort) -> None:
        module_id, port_id = port.kind.module_id, port.kind.port_id
        self.msg = f"This port {module_id}/{port_id} does not support pcs_pma!"
        super().__init__(self.msg)


class NotSupportAutoNegError(ConfigError):
    def __init__(self, port: GenericAnyPort) -> None:
        module_id, port_id = port.kind.module_id, port.kind.port_id
        self.msg = f"This port {module_id}/{port_id} does not support auto negotiation!"
        super().__init__(self.msg)


class NotSupportLinkTrainError(ConfigError):
    def __init__(self, port: GenericAnyPort) -> None:
        module_id, port_id = port.kind.module_id, port.kind.port_id
        self.msg = f"This port {module_id}/{port_id} does not support link training!"
        super().__init__(self.msg)


class NotRightLaneLengthError(ConfigError):
    def __init__(self, serdes: list[int]) -> None:
        self.msg = f"Serdes {serdes} should be length of 4!"
        super().__init__(self.msg)


class NotRightLaneValueError(ConfigError):
    def __init__(self, serdes: list[int]) -> None:
        self.msg = f"Serdes {serdes} should be a list of 4 integers ranges from 0 to 255!"
        super().__init__(self.msg)


class NotSupportMedia(ConfigError):
    def __init__(self, module: GenericAnyModule) -> None:
        module_id = module.module_id
        self.msg = f"This module {module_id} does not support the media configuration!"

    
class NotSupportPortSpeed(ConfigError):
    def __init__(self, module: GenericAnyModule) -> None:
        module_id = module.module_id
        self.msg = f"This module {module_id} does not support the port-speed configuration under its current media configuration!"


__all__ = (
    "ConfigError",
    "NoSuchModuleError",
    "NoSuchPortError",
    "NotConnectedError",
    "NotRightLaneLengthError",
    "NotRightLaneValueError",
    "NotSupportAutoNegError",
    "NotSupportLinkTrainError",
    "NotSupportPcsPmaError",
    "NotSupportMedia",
    "NotSupportPortSpeed",
)



#: Impairment Port Distribution Commands

from dataclasses import dataclass
import functools
import typing

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403


@register_command
@dataclass
class PED_SCHEDULE:
    """
    Configure the impairment scheduler function.  The configuration of the scheduler
    depends on the kind of distribution to schedule:      (1) Burst distributions:
    "Fixed Burst" and "Accumulate and Burst".      (2) Non-Burst distributions: All
    others.  For burst distributions, the scheduler can be configured for "One-shot"
    operation or "Repeat Operation".  When running in "Repeat Operation" the "Repeat
    Period" must be configured. For non-burst distributions,  the scheduler can be
    configured operate in either "Continuous" or "Repeat Period" modes.  When
    running in "Repeat Period" configuration of "Duration" and "Repeat Period" is
    required.
    """

    code: typing.ClassVar[int] = 1611
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        duration: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the "on" period. Units = multiples of 10 ms (range 1 to 65535), default is 1
        period: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the "total" period. Units = multiples of 10 ms (range 0 to 65535), default is 0

    @dataclass(frozen=True)
    class GetDataAttr:
        duration: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the "on" period. Units = multiples of 10 ms (range 1 to 65535), default is 1
        period: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the "total" period. Units = multiples of 10 ms (range 0 to 65535), default is 0

    def get(self) -> "Token[GetDataAttr]":
        """Get the impairment scheduler configuration.

        :return: the impairment scheduler configuration
        :rtype: PED_SCHEDULE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, duration: int, period: int) -> "Token":
        """Set the impairment scheduler configuration.

        :param duration: specifies the "on" period. Units = multiples of 10 ms (range 1 to 65535), default is 1
        :type duration: int
        :param period: specifies the "total" period. Units = multiples of 10 ms (range 0 to 65535), default is 0
        :type period: int
        """
        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], duration=duration, period=period),
        )


@register_command
@dataclass
class PED_ONESHOTSTATUS:
    """
    Retrieves the one-shot completion status.

    .. note::

        The return value is only valid, if the configured distribution is either accumulate & burst (DELAY) or fixed burst (non-DELAY).

    """

    code: typing.ClassVar[int] = 1612
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class GetDataAttr:
        one_shot_status: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, specifies the status.

    def get(self) -> "Token[GetDataAttr]":
        """Get the one-shot completion status.

        :return: the one-shot completion status
        :rtype: PED_ONESHOTSTATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))


@register_command
@dataclass
class PED_OFF:
    """Configure Impairments Distribution to OFF. Assigning a different distribution than OFF to an impairment
    will activate the impairment. To de-activate the impairment assign distribution OFF.
    """

    code: typing.ClassVar[int] = 1620
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Configure Impairments Distribution to OFF. Assigning a different distribution than OFF to an impairment
        will activate the impairment. To de-activate the impairment assign distribution OFF.
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._flow_xindex, self._impairment_type_xindex],
            ),
        )


@register_command
@dataclass
class PED_FIXED:
    """
    Configuration of Fixed Rate distribution. This is predictable distribution with
    nearly equal distance between impairments, to match the configured probability.

    .. note::

        In case of misordering, a special limit applies, probability * (depth + 1) should be less than 1000000.

    """

    code: typing.ClassVar[int] = 1621
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        probability: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the fixed probability in ppm. Default value is 0.

    @dataclass(frozen=True)
    class GetDataAttr:
        probability: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the fixed probability in ppm. Default value is 0.

    def get(self) -> "Token[GetDataAttr]":
        """Get the probability of a Fixed Rate distribution.

        :return: the fixed probability in ppm. Default value is 0.
        :rtype: PED_FIXED.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, probability: int) -> "Token":
        """Set the probability of a Fixed Rate distribution.

        :param probability: the fixed probability in ppm. Default value is 0.
        :type probability: int
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._flow_xindex, self._impairment_type_xindex],
                probability=probability
            )
        )


@register_command
@dataclass
class PED_RANDOM:
    """
    Configuration of Random Rate distribution. Packets are impaired randomly based
    on a per packet probability. This way the impaired fraction of packets will be
    equal to the configured probability over time. Random probability in ppm (i.e. 1
    means 0.0001%)
    """

    code: typing.ClassVar[int] = 1622
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        probability: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the random probability in ppm. Default value is 0.

    @dataclass(frozen=True)
    class GetDataAttr:
        probability: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the random probability in ppm. Default value is 0.

    def get(self) -> "Token[GetDataAttr]":
        """Get the probability of a Random Rate distribution.

        :return: specifies the random probability in ppm. Default value is 0.
        :rtype: PED_RANDOM.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, probability: int) -> "Token":
        """Set the probability of a Random Rate distribution.

        :param probability: specifies the random probability in ppm. Default value is 0.
        :type probability: int
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._flow_xindex, self._impairment_type_xindex],
                probability=probability
            )
        )


@register_command
@dataclass
class PED_BER:
    """
    Configuration of Bit Error Rate distribution.
    """

    code: typing.ClassVar[int] = 1623
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        coef: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the coefficient for BER. Default value: 1 (Range is 1 to 9).
        exp: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the exponent for BER. Default value: -10 (Range is -18 to -1).

    @dataclass(frozen=True)
    class GetDataAttr:
        coef: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the coefficient for BER. Default value: 1 (Range is 1 to 9).
        exp: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the exponent for BER. Default value: -10 (Range is -18 to -1).

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Bit Error Rate distribution.

        :return: the configuration of Bit Error Rate distribution
        :rtype: PED_BER.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, coef: int, exp: int) -> "Token":
        """Set the configuration of Bit Error Rate distribution.

        :param coef: specifies the coefficient for BER. Default value: 1 (Range is 1 to 9).
        :type coef: int
        :param exp: specifies the exponent for BER. Default value: -10 (Range is -18 to -1).
        :type exp: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], coef=coef, exp=exp))


@register_command
@dataclass
class PED_FIXEDBURST:
    """
    Configuration of Fixed Burst distribution.

    .. note::

        In case of ``_impairment_type_xindex`` = ``MISO``, burst size is fixed to 1.

    """

    code: typing.ClassVar[int] = 1624
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        burst_size: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the burst size (Range 1 - 16383). Default value = 1.

    @dataclass(frozen=True)
    class GetDataAttr:
        burst_size: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the burst size (Range 1 - 16383). Default value = 1.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Fixed Burst distribution.

        :return: configuration of Fixed Burst distribution.
        :rtype: PED_FIXEDBURST.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, burst_size: int) -> "Token":
        """Set the configuration of Fixed Burst distribution.

        :param burst_size: specifies the burst size (Range 1 - 16383). Default value = 1.
        :type burst_size: int
        """
        return Token(
            self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], burst_size=burst_size)
        )


@register_command
@dataclass
class PED_RANDOMBURST:
    """
    Configuration of Random Burst distribution.
    """

    code: typing.ClassVar[int] = 1625
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        minimum: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies minimum burst size. Default value: 0 (Range 0 to 65535)
        maximum: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies maximum burst size. Default value: 0 (Range 0 to 65535)
        probability: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the per packet probability of initiating a burst in ppm. Default value: 0.

    @dataclass(frozen=True)
    class GetDataAttr:
        minimum: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies minimum burst size. Default value: 0 (Range 0 to 65535)
        maximum: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies maximum burst size. Default value: 0 (Range 0 to 65535)
        probability: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the per packet probability of initiating a burst in ppm. Default value: 0.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Random Burst distribution.

        :return: configuration of Random Burst distribution.
        :rtype: PED_RANDOMBURST.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, minimum: int, maximum: int, probability: int) -> "Token":
        """Set the configuration of Random Burst distribution.

        :param minimum: specifies minimum burst size. Default value: 0 (Range 0 to 65535)
        :type minimum: int
        :param maximum: specifies maximum burst size. Default value: 0 (Range 0 to 65535)
        :type maximum: int
        :param probability: specifies the per packet probability of initiating a burst in ppm. Default value: 0.
        :type probability: int
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._flow_xindex, self._impairment_type_xindex],
                minimum=minimum,
                maximum=maximum,
                probability=probability
            ),
        )


@register_command
@dataclass
class PED_GE:
    """
    Configuration of Gilbert-Elliot distribution.
    """

    code: typing.ClassVar[int] = 1626
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        good_state_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the good state probability in ppm. Default value: 0.
        good_state_trans_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the good state transition probability in ppm. Default value: 0.
        bad_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the bad state probability in ppm. Default value: 0.
        bad_state_trans_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the bad state transition probability in ppm. Default value: 0.

    @dataclass(frozen=True)
    class GetDataAttr:
        good_state_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the good state probability in ppm. Default value: 0.
        good_state_trans_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the good state transition probability in ppm. Default value: 0.
        bad_state_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the bad state probability in ppm. Default value: 0.
        bad_state_trans_prob: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, specifies the bad state transition probability in ppm. Default value: 0.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Gilbert-Elliot distribution.

        :return: the configuration of Gilbert-Elliot distribution.
        :rtype: PED_GE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, good_state_prob: int, good_state_trans_prob: int, bad_state_prob: int, bad_state_trans_prob: int) -> "Token":
        """Set the configuration of Gilbert-Elliot distribution.

        :param good_state_prob: specifies the good state probability in ppm. Default value: 0.
        :type good_state_prob: int
        :param good_state_trans_prob: specifies the good state transition probability in ppm. Default value: 0.
        :type good_state_trans_prob: int
        :param bad_state_prob: specifies the bad state probability in ppm. Default value: 0.
        :type bad_state_prob: int
        :param bad_state_trans_prob: specifies the bad state transition probability in ppm. Default value: 0.
        :type bad_state_trans_prob: int
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._flow_xindex, self._impairment_type_xindex],
                good_state_prob=good_state_prob,
                good_state_trans_prob=good_state_trans_prob,
                bad_state_prob=bad_state_prob,
                bad_state_trans_prob=bad_state_trans_prob,
            ),
        )


@register_command
@dataclass
class PED_UNI:
    """
    Configuration of Uniform distribution.

    .. note::

        If minimum is less than minimum latency, value is set to minimum latency. If minimum is greater than maximum latency, value is set to maximum latency.

    """

    code: typing.ClassVar[int] = 1627
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        minimum: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, in case of iid != DELAY, specifies the minimum no. of packets. Default value: 0 (Range 0 to 4194288). In case of iid = DELAY, specifies the minimum latency limit. Unit is nanosecond (must be multiples of 100 ns). Default value: minimum latency.
        maximum: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, in case of iid != DELAY, specifies the maximum no. of packets. Default value: 0 (Range 0 to 4194288). In case of iid = DELAY, specifies the maximum latency limit. Unit is nanosecond (must be multiples of 100 ns). Default value: minimum latency.

    @dataclass(frozen=True)
    class GetDataAttr:
        minimum: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, in case of iid != DELAY, specifies the minimum no. of packets. Default value: 0 (Range 0 to 4194288). In case of iid = DELAY, specifies the minimum latency limit. Unit is nanosecond (must be multiples of 100 ns). Default value: minimum latency.
        maximum: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, in case of iid != DELAY, specifies the maximum no. of packets. Default value: 0 (Range 0 to 4194288). In case of iid = DELAY, specifies the maximum latency limit. Unit is nanosecond (must be multiples of 100 ns). Default value: minimum latency.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Uniform distribution.

        :return: the configuration of Uniform distribution.
        :rtype: PED_UNI.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, minimum: int, maximum: int) -> "Token":
        """Set the configuration of Uniform distribution.

        :param minimum: in case of iid != DELAY, specifies the minimum no. of packets. Default value: 0 (Range 0 to 4194288). In case of iid = DELAY, specifies the minimum latency limit. Unit is nanosecond (must be multiples of 100 ns). Default value: minimum latency.
        :type minimum: int
        :param maximum: in case of iid != DELAY, specifies the maximum no. of packets. Default value: 0 (Range 0 to 4194288). In case of iid = DELAY, specifies the maximum latency limit. Unit is nanosecond (must be multiples of 100 ns). Default value: minimum latency.
        :type maximum: int
        """
        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], minimum=minimum, maximum=maximum),
        )


@register_command
@dataclass
class PED_GAUSS:
    """
    Configuration of Gaussian distribution.

    .. note::

        In case of ``_impairment_type_xindex != DELAY``: (1) mean plus 3 times standard deviation should be less than or equal to max allowed (4194288). (2) mean should always be at least 3 times the standard deviation, this to ensure that the impairment distance is always positive.

        In case of ``_impairment_type_xindex = DELAY``: (1) mean plus 3 times standard deviation should be less than or equal to the maximum latency. (2) mean minus 3 times the standard deviation should be greater than or equal to minimum latency.

    """

    code: typing.ClassVar[int] = 1628
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mean: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Gaussian mean. In case of iid != DELAY, specifies the Gaussian mean value as number of packets.Default value: 0 packets (Range 0 to 4194288). In case of iid = DELAY, specifies the Gaussian mean value. Units is nanosecond (must be multiples of 100 ns).
        std_deviation: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Gaussian standard deviation. In case of iid != DELAY, specifies the standard deviation as number of packets. Default value: 0 packets (Range 0 to 4194288). In case of iid = DELAY, specifies the the Gaussian standard deviation. Units is nanosecond (must be multiples of 100 ns). Default value: 0 ns.

    @dataclass(frozen=True)
    class GetDataAttr:
        mean: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Gaussian mean. In case of iid != DELAY, specifies the Gaussian mean value as number of packets.Default value: 0 packets (Range 0 to 4194288). In case of iid = DELAY, specifies the Gaussian mean value. Units is nanosecond (must be multiples of 100 ns).
        std_deviation: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Gaussian standard deviation. In case of iid != DELAY, specifies the standard deviation as number of packets. Default value: 0 packets (Range 0 to 4194288). In case of iid = DELAY, specifies the the Gaussian standard deviation. Units is nanosecond (must be multiples of 100 ns). Default value: 0 ns.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Gaussian distribution.

        :return: the configuration of Gaussian distribution
        :rtype: PED_GAUSS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, mean: int, std_deviation: int) -> "Token":
        """Set the configuration of Gaussian distribution.

        :param mean: specifies the Gaussian mean.
        :type mean: int
        :param std_deviation: specifies the Gaussian standard deviation.
        :type std_deviation: int
        """
        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], mean=mean, std_deviation=std_deviation),
        )


@register_command
@dataclass
class PED_POISSON:
    """
    Configuration of "Poisson" distribution.

    .. note::

        Standard deviation is derived from mean, i.e., standard deviation = SQRT(mean).

        In case of ``_impairment_type_xindex != DELAY``, mean plus 3 times standard deviation should be less than or equal to max allowed (4194288).

        In case of ``_impairment_type_xindex = DELAY``, mean plus 3 times standard deviation should be less than or equal to the maximum latency.

    """

    code: typing.ClassVar[int] = 1629
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mean: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Poisson mean value. In case of iid = DELAY specifies the Poisson mean. Unit is nanosecond (must be multiples of 100ns). Default value: 0 ns. In case of iid != DELAY specifies the Poisson mean in number of packets packets. Default value: 9 packets (Range 0 to 4194288).

    @dataclass(frozen=True)
    class GetDataAttr:
        mean: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Poisson mean value. In case of iid = DELAY specifies the Poisson mean. Unit is nanosecond (must be multiples of 100ns). Default value: 0 ns. In case of iid != DELAY specifies the Poisson mean in number of packets packets. Default value: 9 packets (Range 0 to 4194288).

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Poisson distribution.

        :return: the configuration of Poisson distribution
        :rtype: PED_POISSON.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, mean: int) -> "Token":
        """Set the configuration of Poisson distribution.

        :param mean: specifies the Poisson mean value.
        :type mean: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], mean=mean))


@register_command
@dataclass
class PED_GAMMA:
    """
    Configuration of Gamma distribution.

    .. note::

        Mean and Standard deviation are calculated from Shape and Scale parameters and validation is performed using those. standard deviation = [SQRT(shape * scale * scale)]mean = [shape * scale].

        In case of ``_impairment_type_xindex != DELAY``, (1) mean plus 4 times standard deviation should be less than or equal to max allowed(4194288). (2)shape and scale should be greater than or equal to 0.

        In case of ``_impairment_type_xindex = DELAY``, mean plus 4 times standard deviation should be less than or equal to the maximum latency.

    """

    code: typing.ClassVar[int] = 1630
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        shape: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long, specifies the shape. Units: none. Default value: 0.
        scale: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Gamma function scale parameter. In case of iid = DELAY, units: nanosecond (must be multiples of 100 ns). Default value: 0 ns. In case of iid != DELAY, units: number of packets.Default value: 0 packets.

    @dataclass(frozen=True)
    class GetDataAttr:
        shape: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long, specifies the shape. Units: none. Default value: 0.
        scale: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the Gamma function scale parameter. In case of iid = DELAY, units: nanosecond (must be multiples of 100 ns). Default value: 0 ns. In case of iid != DELAY, units: number of packets.Default value: 0 packets.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Gamma distribution.

        :return: the configuration of Gamma distribution
        :rtype: PED_GAMMA.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, shape: int, scale: int) -> "Token":
        """Set the configuration of Gamma distribution.

        :param shape: specifies the shape. Units: none. Default value: 0.
        :type shape: int
        :param scale: specifies the Gamma function scale parameter.
        :type scale: int
        """
        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], shape=shape, scale=scale),
        )


@register_command
@dataclass
class PED_CUST:
    """
    Associate a custom distribution to a flow and impairment type.

    .. note::

        Before associating a custom distribution, the below validation checks are applied.

        In case of ``_impairment_type_xindex != DELAY``, (1) Custom values should be less than or equal to max allowed (4194288). (2) Custom distribution bust contain 512 values.

        In case of ``_impairment_type_xindex = DELAY``, (1) Custom values should be less than or equal to the maximum latency. (2) Custom values should be greater than or equal to minimum latency. (3) Custom distribution should contain 1024 values.

    """

    code: typing.ClassVar[int] = 1631
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        cust_id: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, custom distribution identifier.

    @dataclass(frozen=True)
    class GetDataAttr:
        cust_id: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, custom distribution identifier.

    def get(self) -> "Token[GetDataAttr]":
        """Get the custom distribution identifier that is associated to a flow and impairment type.

        :return: custom distribution identifier
        :rtype: PED_CUST.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, cust_id: int) -> "Token":
        """Associate a custom distribution to a flow and impairment type.

        :param cust_id: custom distribution identifier
        :type cust_id: int
        """
        return Token(
            self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], cust_id=cust_id)
        )


@register_command
@dataclass
class PED_CONST:
    """
    Configuration of Constant Delay distribution (DELAY only). Unit is ns (must be
    multiples of 100ns). Default value: Minimum supported per speed and FEC mode.

    .. note::

        If the latency is less than minimum latency, value is set to minimum latency. If the latency is greater than maximum latency, value is set to maximum latency.

    """

    code: typing.ClassVar[int] = 1640
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        delay: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the constant delay/latency time. Unit is nanosecond (must be multiples of 100 ns). Default value: Minimum supported per speed and FEC mode.

    @dataclass(frozen=True)
    class GetDataAttr:
        delay: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the constant delay/latency time. Unit is nanosecond (must be multiples of 100 ns). Default value: Minimum supported per speed and FEC mode.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Constant Delay distribution (DELAY only).

        :return: the configuration of Constant Delay distribution (DELAY only)
        :rtype: PED_CONST.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, delay: int) -> "Token":
        """Set the configuration of Constant Delay distribution (DELAY only).

        :param delay: specifies the constant delay/latency time. Unit is nanosecond (must be multiples of 100 ns). Default value: Minimum supported per speed and FEC mode.
        :type delay: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], delay=delay))


@register_command
@dataclass
class PED_ACCBURST:
    """
    Configuration of Accumulate & Burst distribution (DELAY only).

    .. note::

        If the delay is less than minimum latency, value is set to minimum latency. If the delay is greater than maximum latency, value is set to maximum latency.

    """

    code: typing.ClassVar[int] = 1641
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        delay: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the burst delay time. Units = nanosecond (must multiples of 100 ns). Default value: minimum latency.

    @dataclass(frozen=True)
    class GetDataAttr:
        delay: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long, specifies the burst delay time. Units = nanosecond (must multiples of 100 ns). Default value: minimum latency.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Accumulate & Burst distribution (DELAY only).

        :return: the configuration of Accumulate & Burst distribution (DELAY only)
        :rtype: PED_ACCBURST.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, delay: int) -> "Token":
        """Set the configuration of Accumulate & Burst distribution (DELAY only).

        :param delay: specifies the burst delay time. Units = nanosecond (must multiples of 100 ns). Default value: minimum latency.
        :type delay: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], delay=delay))


@register_command
@dataclass
class PED_STEP:
    """
    Configuration of Step distribution (DELAY only).

    .. note::

        If the low/high is less than minimum latency, value is set to minimum latency. If the low/high is greater than maximum latency, value is set to maximum latency.

    """

    code: typing.ClassVar[int] = 1642
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        low: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long, specifies the packet delay in the 'low' state of the step. Units = nanosecond (must be multiples of 100 ns).
        high: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long, specifies the packet delay in the 'high' state of the step. Units = nanosecond (must be multiples of 100 ns).

    @dataclass(frozen=True)
    class GetDataAttr:
        low: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long, specifies the packet delay in the 'low' state of the step. Units = nanosecond (must be multiples of 100 ns).
        high: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long, specifies the packet delay in the 'high' state of the step. Units = nanosecond (must be multiples of 100 ns).

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of Step distribution (DELAY only).

        :return: the configuration of Step distribution (DELAY only)
        :rtype: PED_STEP.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, low: int, high: int) -> "Token":
        """Set the configuration of Step distribution (DELAY only).

        :param low: specifies the packet delay in the 'low' state of the step. Units = nanosecond (must be multiples of 100 ns).
        :type low: int
        :param high: specifies the packet delay in the 'high' state of the step. Units = nanosecond (must be multiples of 100 ns).
        :type high: int
        """
        return Token(
            self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], low=low, high=high)
        )


# TODO: add descriptions
@register_command
@dataclass
class PED_ENABLE:
    """
    Control whether this impairment distribution is enabled.

    .. note::

        This command is not applicable for PE_BANDPOLICER and PE_BANDSHAPER because they have a separate ``ON / OFF`` parameter.

    """

    code: typing.ClassVar[int] = 1644
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _flow_xindex: int
    _impairment_type_xindex: int

    @dataclass(frozen=True)
    class GetDataAttr:
        action: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, specifying whether impairment is enabled.

    @dataclass(frozen=True)
    class SetDataAttr:
        action: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, specifying whether impairment is enabled.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of this impairment distribution.

        :return: the status of this impairment distribution
        :rtype: PED_ENABLE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex]))

    def set(self, action: OnOff) -> "Token[GetDataAttr]":
        """Set the status of this impairment distribution.

        :param action: the status of this impairment distribution
        :type action: OnOff
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._impairment_type_xindex], action=action))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable impairment distribution"""
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable impairment distribution"""



import asyncio
from typing import (
    Type,
    TypeVar,
    Protocol,
    TYPE_CHECKING,
)
from collections import OrderedDict
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from ._base_manager import ResourcesBaseManager


class IPort(Protocol):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        ...


PT = TypeVar("PT", bound=IPort)


class PortsManager(ResourcesBaseManager[PT]):

    __slots__ = ("_conn", "_ports_type", "_ports_count", "_module_id", )

    def __init__(self, conn: "itf.IConnection", module_id: int, ports_type: Type[PT], ports_count: int) -> None:
        super().__init__()
        self._conn = conn
        self._ports_type = ports_type
        self._ports_count = ports_count
        self._module_id = module_id
        self._items: OrderedDict[int, PT] = OrderedDict(
            (
                port_id,
                self._ports_type(
                    conn=self._conn,
                    module_id=self._module_id,
                    port_id=port_id,
                )
            )
            for port_id in range(self._ports_count)
        )

    async def fill(self) -> None:
        """Method for create and fill in."""

        assert not self._lock, "Method <fill> can be called only once."
        coros = list(self._items.values())
        await asyncio.gather(*coros)  # type: ignore


class PortResolver(Protocol):
    async def __call__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> Type:
        ...


class PortsCombiManager(ResourcesBaseManager[PT]):

    __slots__ = ("_conn", "_resolver", "_ports_count", "_module_id", )

    def __init__(self, conn: "itf.IConnection", module_id: int, resolver: PortResolver, ports_count: int) -> None:
        super().__init__()
        self._conn = conn
        self._resolver = resolver
        self._ports_count = ports_count
        self._module_id = module_id
        self._items: OrderedDict[int, PT] = OrderedDict()

    async def fill(self) -> None:
        """Method for create and fill in."""
        assert not self._lock, "Method <fill> can be called only once."
        coros = [
            self._resolver(
                conn=self._conn,
                module_id=self._module_id,
                port_id=port_id,
            )
            for port_id in range(self._ports_count)
        ]
        ports = await asyncio.gather(*coros)
        self._items = OrderedDict(
            (port.kind.port_id, port) for port in ports
        )



# : L23 Port Layer-1 Commands

from dataclasses import dataclass
import typing
import functools

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403


@register_command
@dataclass
class PL1_AUTONEGINFO:
    """
    .. versionadded:: 1.1

    .. warning::

        Still in beta mode. Subjected to changes

    Get L1 auto-negotiation information. Information is split into a number of pages.

    """

    code: typing.ClassVar[int] = 385
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _page_xindex: int

    @dataclass(frozen=True)
    class GetDataAttr:

        rx_link_codeword_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # received number of Link Code Words (Base Pages).

        rx_next_page_message_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # received number of Next Pages - Message Pages.

        rx_next_page_unformatted_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # received number of Nex Pages - Unformatted Pages.

        tx_link_codeword_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # transmitted number of Link Code Words (Base Pages).

        tx_next_page_message_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # transmitted number of Next Pages - Message Pages.

        tx_next_page_unformatted_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # transmitted number of Nex Pages - Unformatted Pages.

        negotiation_hcd_fail_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # number of negotiation HCD (Highest Common Denominator) failures.

        negotiation_fec_fail_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # number of negotiation FEC failures.

        negotiation_loss_of_sync_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # number of negotiation Loss of Sync failures.

        negotiation_timeout_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # number of negotiation timeouts.

        negotiation_success_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # number of negotiation successes.

        duration_us: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # duration of the auto-negotiation in microseconds, from autoneg is enabled on the port to the negotiation is finished.

    def get(self) -> "Token[GetDataAttr]":
        """Get L1 auto-negotiation information. Information is split into a number of pages.

        :return: L1 auto-negotiation information
        :rtype: PL1_AUTONEGINFO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._page_xindex]))


@register_command
@dataclass
class PL1_LINKTRAININFO:
    """
    .. versionadded:: 1.1

    .. warning::

        Still in beta mode. Subjected to changes

    Get L1 link training information. Information is per Serdes and split into a number of pages.
    """

    code: typing.ClassVar[int] = 386
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _serdes_xindex: int
    _page_xindex: int

    @dataclass(frozen=True)
    class GetDataAttr:

        duration_us: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # duration of the link training process in microseconds, from LT is enabled on the port to the LT is finished.

        lock_lost_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # number of lost locks on auto-neg.

        pre1_current_level: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) current level.

        pre1_rx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) received number of increment requests.

        pre1_rx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) received number of decrement requests.

        pre1_rx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) received number of maximum limits of coefficient and equalization requests reached.

        pre1_rx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) received number of maximum limits of equalization requests reached.

        pre1_rx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) received number of coefficients not supported.

        pre1_rx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) received number of coefficients at limit.

        pre1_tx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) transmitted number of increment requests.

        pre1_tx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) transmitted number of decrement requests.

        pre1_tx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) transmitted number of maximum limits of coefficient and equalization requests reached.

        pre1_tx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) transmitted number of maximum limits of equalization requests reached.

        pre1_tx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) transmitted number of coefficients not supported.

        pre1_tx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-1) transmitted number of coefficients at limit.

        main_current_level: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) current level.

        main_rx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) received number of increment requests.

        main_rx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) received number of decrement requests.

        main_rx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) received number of maximum limits of coefficient and equalization requests reached.

        main_rx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) received number of maximum limits of equalization requests reached.

        main_rx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) received number of coefficients not supported.

        main_rx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) received number of coefficients at limit.

        main_tx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) transmitted number of increment requests.

        main_tx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) transmitted number of decrement requests.

        main_tx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) transmitted number of maximum limits of coefficient and equalization requests reached.

        main_tx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) transmitted number of maximum limits of equalization requests reached.

        main_tx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) transmitted number of coefficients not supported.

        main_tx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(0) transmitted number of coefficients at limit.

        post1_current_level: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) current level.

        post1_rx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) received number of increment requests.

        post1_rx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) received number of decrement requests.

        post1_rx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) received number of maximum limits of coefficient and equalization requests reached.

        post1_rx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) received number of maximum limits of equalization requests reached.

        post1_rx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) received number of coefficients not supported.

        post1_rx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) received number of coefficients at limit.

        post1_tx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) transmitted number of increment requests.

        post1_tx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) transmitted number of decrement requests.

        post1_tx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) transmitted number of maximum limits of coefficient and equalization requests reached.

        post1_tx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) transmitted number of maximum limits of equalization requests reached.

        post1_tx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) transmitted number of coefficients not supported.

        post1_tx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(1) transmitted number of coefficients at limit.

        pre2_current_level: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) current level.

        pre2_rx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) received number of increment requests.

        pre2_rx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) received number of decrement requests.

        pre2_rx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) received number of maximum limits of coefficient and equalization requests reached.

        pre2_rx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) received number of maximum limits of equalization requests reached.

        pre2_rx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) received number of coefficients not supported.

        pre2_rx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) received number of coefficients at limit.

        pre2_tx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) transmitted number of increment requests.

        pre2_tx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) transmitted number of decrement requests.

        pre2_tx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) transmitted number of maximum limits of coefficient and equalization requests reached.

        pre2_tx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) transmitted number of maximum limits of equalization requests reached.

        pre2_tx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) transmitted number of coefficients not supported.

        pre2_tx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-2) transmitted number of coefficients at limit.

        pre3_current_level: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) current level.

        pre3_rx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) received number of increment requests.

        pre3_rx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) received number of decrement requests.

        pre3_rx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) received number of maximum limits of coefficient and equalization requests reached.

        pre3_rx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) received number of maximum limits of equalization requests reached.

        pre3_rx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) received number of coefficients not supported.

        pre3_rx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) received number of coefficients at limit.

        pre3_tx_increment_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) transmitted number of increment requests.

        pre3_tx_decrement_req_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) transmitted number of decrement requests.

        pre3_tx_coeff_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) transmitted number of maximum limits of coefficient and equalization requests reached.

        pre3_tx_eq_limit_reached_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) transmitted number of maximum limits of equalization requests reached.

        pre3_tx_coeff_not_supported_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) transmitted number of coefficients not supported.

        pre3_tx_coeff_at_limit_count: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # c(-3) transmitted number of coefficients at limit.

        prbs_total_bits_high: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # PRBS total bits (most significant 32-bit).

        prbs_total_bits_low: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # PRBS total bits  (least significant 32-bit).

        prbs_total_error_bits_high: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # PRBS total error bits (most significant 32-bit, only bit 15-0 should be used).

        prbs_total_error_bits_low: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # PRBS total error bits (least significant 32-bit).

        frame_lock: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt, choices=LinkTrainFrameLock)  # frame lock status of the local end.

        remote_frame_lock: XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt, choices=LinkTrainFrameLock)  # frame lock status of the remote end.

        num_frame_errors:  XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # Number of frame errors received

        num_overruns:  XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # Number of overruns

        last_ic_received:  XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # Last preset request received

        last_ic_sent:  XmpField[xt.XmpUnsignedInt] = XmpField(xt.XmpUnsignedInt)  # Last preset request sent

    def get(self) -> "Token[GetDataAttr]":
        """Get L1 link training information. Information is per Serdes and split into a number of pages.

        :return: L1 link training information
        :rtype: PL1_LINKTRAININFO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._page_xindex]))


@register_command
@dataclass
class PL1_LOG:
    """
    .. versionadded:: 1.1

    Return ANLT log messages in json format.
    """

    code: typing.ClassVar[int] = 387
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class GetDataAttr:

        log_string: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)

    def get(self) -> "Token[GetDataAttr]":
        """Return anlt log in json format

        :return: anlt log in json format
        :rtype: PL1_LOG.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PL1_CFG_TMP:
    """
    .. versionadded:: 1.1

    .. warning::

        Still in beta mode. Subjected to changes

    Configure some L1 parameters.
    """

    code: typing.ClassVar[int] = 388
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _serdes_xindex: int
    _type: int

    @dataclass(frozen=True)
    class GetDataAttr:
        """Data structure of the get response.
        """
        values: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)

    @dataclass(frozen=True)
    class SetDataAttr:
        """Data structure of the set action.
        """
        values: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)

    def get(self) -> "Token[GetDataAttr]":
        """Get various L1 parameters

        :return: various L1 parameters
        :rtype: PL1_CFG_TMP.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._type]))

    def set(self, values: typing.List[int]) -> "Token":
        """Get various L1 parameters

        :param values: L1 parameters
        :type values: typing.List[int]
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._type], values=values))


@register_command
@dataclass
class PL1_LINKTRAIN_CMD:
    """
    .. versionadded:: 1.1

    .. warning::

        Still in beta mode. Subjected to changes

    Link training RPC. Issue link training commands on a given serdes and poll for status
    """

    code: typing.ClassVar[int] = 389
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _serdes_xindex: int

    @dataclass(frozen=True)
    class GetDataAttr:
        """Data structure of the get response.
        """
        cmd: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)
        arg: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)
        result: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=LinkTrainCmdResults)
        flags: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)

    @dataclass(frozen=True)
    class SetDataAttr:
        """Data structure of the set action.
        """
        cmd: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=LinkTrainCmd)
        arg: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)

    def get(self) -> "Token[GetDataAttr]":
        """Get status of current command

        :return: 4 bytes: command, arg, result, flags
        :rtype: PL1_LINKTRAIN_CMD.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex]))

    def set(self, cmd, arg) -> "Token":
        """Issue a link train command (cmd, arg)

        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex], cmd=cmd, arg=arg))



import asyncio
from typing import List
from dataclasses import (
    dataclass,
    field,
)
from xoa_driver.internals.core.transporter import funcs
from xoa_driver.internals.core.commands import enums
from xoa_driver.internals.core.commands import (
    M_MODEL,
    M_RESERVATION,
    M_RESERVEDBY,
    M_MEDIASUPPORT,
)

from xoa_driver.internals.utils import attributes as utils


class ModuleLocalState:
    """Module local state.
    """
    __slots__ = (
        "reservation",
        "reserved_by",
        "model",
    )

    def __init__(self) -> None:
        self.reservation: enums.ReservedStatus = enums.ReservedStatus.RELEASED
        self.reserved_by: str = ""
        self.model: str = ""

    async def initiate(self, module) -> None:
        (
            reservation_r,
            reserved_by_r,
            model_r,
        ) = await funcs.apply(
            module.reservation.get(),
            module.reserved_by.get(),
            module.model.get(),
        )
        self.reservation = enums.ReservedStatus(reservation_r.operation)
        self.reserved_by = reserved_by_r.username
        self.model = model_r.model

    def register_subscriptions(self, module) -> None:
        module._conn.subscribe(M_RESERVEDBY, utils.Update(self, "reserved_by", "username", module._check_identity))
        module._conn.subscribe(M_RESERVATION, utils.Update(self, "reservation", "operation", module._check_identity, format=lambda a: enums.ReservedStatus(a)))
        module._conn.subscribe(M_MODEL, utils.Update(self, "model", "model", module._check_identity))


@dataclass(frozen=True)
class ModuleSpeed:
    """Module's port-speed information.
    """
    port_count: int
    """Port count

    :return: number of ports that have the same speed
    :rtype: int
    """

    port_speed: int
    """Port speed

    :return: speed of the ports
    :rtype: int
    """


@dataclass(frozen=True)
class MediaInfo:
    """Module media information
    """
    cage_type: "enums.MediaConfigurationType"
    """Module Media Configuration

    :return: module media configuration
    :rtype: MediaConfigurationType
    """

    available_speeds: List["ModuleSpeed"] = field(default_factory=list)
    """List of module's port-speed information

    :return: list of module's port-speed information
    :rtype: List[ModuleSpeed]
    """


class ModuleL23LocalState(ModuleLocalState):
    """L23 Module local state
    """
    __slots__ = ("__media_info_list",)

    def __init__(self) -> None:
        self.__media_info_list: List["MediaInfo"] = []

    @property
    def media_info_list(self) -> List["MediaInfo"]:
        return self.__media_info_list

    @media_info_list.setter
    def media_info_list(self, value: List[int]) -> None:
        self.__media_info_list.clear()
        _vs = value[:]
        while _vs:
            cage_type = enums.MediaConfigurationType(_vs.pop(0))
            available_speeds_count = _vs.pop(0)
            mi = MediaInfo(
                cage_type,
                [
                    ModuleSpeed(_vs.pop(0), _vs.pop(0))
                    for _ in range(available_speeds_count)
                ]
            )
            self.__media_info_list.append(mi)

    async def initiate(self, module) -> None:
        m_support_resp, *_ = await asyncio.gather(
            M_MEDIASUPPORT(module._conn, module.module_id).get(),
            super().initiate(module)
        )
        self.media_info_list = m_support_resp.media_info_list  # type: ignore

    def register_subscriptions(self, module) -> None:
        super().register_subscriptions(module)
        module._conn.subscribe(M_MEDIASUPPORT, utils.Update(self, "media_info_list", "media_info_list", module._check_identity))



import re
import struct
from abc import ABCMeta
from collections import UserList
from ipaddress import IPv4Address, IPv6Address
from typing import (
    Any,
    List,
    Optional,
    Type,
    Union,
    Generic,
)
from . import exceptions as excp
from . import interfaces as itf

__all__ = (
    "XmpEmpty",
    "XmpIPV6Address",
    "XmpIPV4Address",
    "XmpByte",
    "XmpLong",
    "XmpInt",
    "XmpUnsignedInt",
    "XmpShort",
    "XmpStr",
    "XmpMacAddress",
    "XmpHex",
    "XmpHex1",
    "XmpHex2",
    "XmpHex3",
    "XmpHex4",
    "XmpHex6",
    "XmpHex8",
    "XmpHex16",
    "XmpByteList",
    "XmpIntList",
    "XmpIPV4AddressList",
    "XmpLongList",
    "XmpShortList",
    "XmpHexList",
    "XmpIntList5",
    "XmpIntList10",
    "XmpLongListStopToKeep8",
    "XmpIPV4AddressListStopToKeep4",
)


class XmpEmpty:
    size: int = 0

    def byte_length(self) -> int:
        return self.size


class XmpIPV6Address(IPv6Address):
    """Type for representing Ipv6 addresses"""

    size: int = 16
    # format: str = "q"

    def __bytes__(self) -> bytes:
        return self.packed

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(data)


class XmpIPV4Address(IPv4Address):
    """Type for representing Ipv4 addresses"""

    size: int = 4
    # format: str = "I"

    def __bytes__(self) -> bytes:
        return self.packed

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(data)


class XmpByte(int):
    size: int = 1
    # format: str = "b"
    _max_val = 0xFF
    _max_length = _max_val.bit_length()

    def __new__(cls, *args, **kwargs):
        value = super().__new__(cls, *args, **kwargs)
        if not value.bit_length() <= cls._max_length:
            raise excp.NumberRangeError(value, cls._max_val)
        return value

    def __bytes__(self) -> bytes:
        return self.to_bytes(self.size, "big", signed=False)

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(int.from_bytes(data, "big", signed=False))


class XmpLong(int):
    size: int = 8
    # format: str = "q"
    _max_val = 0xFFFFFFFFFFFFFFFF
    _max_length = _max_val.bit_length()

    def __new__(cls, *args, **kwargs):
        value = super().__new__(cls, *args, **kwargs)
        if not value.bit_length() <= cls._max_length:
            raise excp.NumberRangeError(value, cls._max_val)
        return value

    def __bytes__(self) -> bytes:
        return self.to_bytes(self.size, "big", signed=True)

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(int.from_bytes(data, "big", signed=True))


class XmpInt(int):
    size: int = 4
    # format: str = "i"
    _max_val = 0xFFFFFFFF
    _max_length = _max_val.bit_length()

    def __new__(cls, *args, **kwargs):
        value = super().__new__(cls, *args, **kwargs)
        if not value.bit_length() <= cls._max_length:
            raise excp.NumberRangeError(value, cls._max_val)
        return value

    def __bytes__(self) -> bytes:
        # return pack("!i", self)
        return self.to_bytes(self.size, "big", signed=True)

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(int.from_bytes(data, "big", signed=True))
        # return cls(struct.unpack("!i", data)[0])
        # return cls(ResponseConverter.to_py_signed_int(data))


class XmpUnsignedInt(int):
    size: int = 4
    # format: str = "i"
    _max_val = 0xFFFFFFFF
    _max_length = _max_val.bit_length()

    def __new__(cls, *args, **kwargs):
        value = super().__new__(cls, *args, **kwargs)
        if not value.bit_length() <= cls._max_length:
            raise excp.NumberRangeError(value, cls._max_val)
        return value

    def __bytes__(self) -> bytes:
        # return pack("!i", self)
        return self.to_bytes(self.size, "big", signed=False)

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(int.from_bytes(data, "big", signed=False))


class XmpShort(int):
    size: int = 2
    # format: str = "h"
    _max_val = 0xFFFF
    _max_length = _max_val.bit_length()

    def __new__(cls, *args, **kwargs):
        value = super().__new__(cls, *args, **kwargs)
        if not value.bit_length() <= cls._max_length:
            raise excp.NumberRangeError(value, cls._max_val)
        return value

    def __bytes__(self) -> bytes:
        return self.to_bytes(self.size, "big", signed=True)

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(int.from_bytes(data, "big", signed=True))


class XmpStr(str):
    size: int = 0
    # format: str = "${size}s"

    def __bytes__(self) -> bytes:
        return struct.pack(f"!{len(self)}s", self.encode())

    def byte_length(self) -> int:
        return self.__len__()

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(data.decode())


class XmpMacAddress(str):
    size: int = 6
    # format: str = "6s"

    def __new__(cls, val: Union[int, str]):
        _val = f"{val:012x}" if isinstance(val, int) else val
        tbl = _val.maketrans("", "", ":.- ")
        clean_val = _val.translate(tbl)

        if len(clean_val) != 12 or any(
            (i not in "1234567890ABCDEFabcdef" for i in clean_val)
        ):
            raise excp.InvalidMacAddressError(val)
        return super().__new__(cls, clean_val.upper())

    def __bytes__(self) -> bytes:
        return bytes.fromhex(self)

    def __str__(self) -> str:
        return bytes.fromhex(self).hex(":")

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(data.hex(":"))

    def byte_length(self) -> int:
        return self.size


class XmpHex(str):
    size: int = 1
    # format: str = f"{size}s"

    def __new__(cls, val: str):
        value = super().__new__(cls, val)
        if not value.startswith(("0x", "0X")):
            raise excp.StartWithError(value, "0x", "0X")
        real_val = value.replace("0X", "").replace("0x", "")

        val_value = int.from_bytes(bytes.fromhex(real_val), "big")
        max_val = (1 << (cls.size * 8)) - 1
        if not 0 <= val_value <= max_val:
            raise excp.HexRangeError(value, max_val)
        return value

    def __bytes__(self) -> bytes:
        return bytes.fromhex(self.replace("0X", "").replace("0x", ""))

    def byte_length(self) -> int:
        return self.size

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericType"], data: bytes) -> "itf.XmpGenericType":
        return cls(f"0x{data.hex().upper()}")


class XmpHex1(XmpHex):
    size = 1


class XmpHex2(XmpHex):
    size = 2


class XmpHex3(XmpHex):
    size = 3


class XmpHex4(XmpHex):
    size = 4


class XmpHex6(XmpHex):
    size = 6


class XmpHex8(XmpHex):
    size = 8


class XmpHex16(XmpHex):
    size = 16


class __MegaABCAttrChecking(ABCMeta):
    def __new__(cls, *args, **kwargs):
        inst = super().__new__(cls, *args, **kwargs)
        if args[0] != "XmpDefaultList" and issubclass(
            getattr(inst, "element_type", NotImplementedError), NotImplementedError
        ):
            raise excp.ElementTypeError(args[0])
        return inst


# Complex types
class XmpDefaultList(UserList, Generic[itf.XmpGenericList], metaclass=__MegaABCAttrChecking):
    element_type: Type[itf.IXmpType]
    fix_length: Optional[int] = None
    stop_to_keep: Optional[int] = None
    size: int = 0

    def __init__(self, data: Optional[List[Any]] = None) -> None:
        if not data:
            self.data = []
        else:
            if getattr(self.element_type, "is_add_on", False):
                self.data = data
            else:
                allowed_types = (list, tuple, XmpDefaultList, str)
                if not isinstance(data, allowed_types):
                    raise excp.NotAllowedTypeError(data, allowed_types)

                if isinstance(data, str):
                    data = [
                        "0x" + i
                        for i in re.findall(
                            r".{2}", data.replace("0x", "").replace("0x", "")
                        )
                    ]
                self.data = [self.element_type(el) for el in data]
        if self.fix_length and self.fix_length > 0 and len(self.data) != self.fix_length:
            raise excp.FixedLenghtError(self.data, self.fix_length)

    def __setitem__(self, idx: int, val: Any) -> None:
        self.data[idx] = self.element_type(val)

    def __bytes__(self) -> bytes:
        return b"".join(bytes(el) for el in self.data)

    def insert(self, idx, val) -> None:
        super().insert(idx, self.element_type(val))

    def append(self, val) -> None:
        super().append(self.element_type(val))

    @classmethod
    def from_bytes(cls: Type["itf.XmpGenericList"], data: bytes) -> "itf.XmpGenericList":
        pointer = 0
        each_size = cls.element_type.size
        times: int = (
            cls.fix_length
            if cls.fix_length and cls.fix_length > 0
            else len(data) // cls.element_type.size
        )
        result = []
        for _ in range(times):
            each_ele = cls.element_type.from_bytes(data[pointer: pointer + each_size])
            result.append(each_ele)
            pointer += each_size

        return cls(result)

    def byte_length(self) -> int:
        return self.__len__() * self.element_type.size


class XmpByteList(XmpDefaultList):
    element_type = XmpByte


class XmpIntList(XmpDefaultList):
    element_type = XmpInt


class XmpIPV4AddressList(XmpDefaultList):
    element_type = XmpIPV4Address


class XmpLongList(XmpDefaultList):
    element_type = XmpLong


class XmpShortList(XmpDefaultList):
    element_type = XmpShort


class XmpHexList(XmpDefaultList):
    element_type = XmpHex


class XmpIntList5(XmpIntList):
    fix_length = 5


class XmpIntList10(XmpIntList):
    fix_length = 10


class XmpLongListStopToKeep8(XmpLongList):
    stop_to_keep = XmpLong.size


class XmpIPV4AddressListStopToKeep4(XmpIPV4AddressList):
    stop_to_keep = 4



class WrongTesterError(Exception):
    def __init__(self) -> None:
        self.msg = "Tester you are trying to connect is of a different type."
        super().__init__(self.msg)


class WrongTesterPasswordError(Exception):
    def __init__(self, password: str) -> None:
        self.password = password
        self.msg = "Can't create session on the tester because the password is incorrect."
        super().__init__(self.msg)



"""
L47 Module Packet Engine Commands
"""
from dataclasses import dataclass
import typing
import functools

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403


@register_command
@dataclass
class M4E_MODE:
    """
    Select resource allocation mode.
    """

    code: typing.ClassVar[int] = 850
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mode: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=ResourceAllocationMode)  # coded byte, resource allocation mode.

    @dataclass(frozen=True)
    class GetDataAttr:
        mode: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=ResourceAllocationMode)  # coded byte, resource allocation mode.

    def get(self) -> "Token[GetDataAttr]":
        """Get the resource allocation mode.

        :return: resource allocation mode
        :rtype: M4E_MODE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, mode: ResourceAllocationMode) -> "Token":
        """Set the resource allocation mode.

        :param mode: resource allocation mode
        :type mode: ResourceAllocationMode
        """
        return Token(self._connection, build_set_request(self, module=self._module, mode=mode))

    set_simple = functools.partialmethod(set, ResourceAllocationMode.SIMPLE)
    """Set resource allocation mode to Simple."""
    set_advanced = functools.partialmethod(set, ResourceAllocationMode.ADVANCED)
    """Set resource allocation mode to Advanced."""


@register_command
@dataclass
class M4E_RESERVE:
    """
    Advanced mode only: Reserve a number of PEs so they later can be assigned to specific ports.
    """

    code: typing.ClassVar[int] = 851
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mask: XmpField[xt.XmpHex8] = XmpField(xt.XmpHex8)  # eight hex bytes, bitmask of PEs to reserve

    @dataclass(frozen=True)
    class GetDataAttr:
        mask: XmpField[xt.XmpHex8] = XmpField(xt.XmpHex8)  # eight hex bytes, bitmask of PEs to reserve

    def get(self) -> "Token[GetDataAttr]":
        """Get the PEs reserved.

        :return: the number of PEs reserved.
        :rtype: M4E_RESERVE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, mask: str) -> "Token":
        """Set PEs reserved.

        :param mask: bitmask of PEs to reserve
        :type mask: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, mask=mask))
